@@ -4,6 +4,9 @@ from django.core.urlresolvers import resolve
 from django.contrib.auth import authenticate
 from django.utils import simplejson
 from django.contrib.auth.models import AnonymousUser
+from django.db.utils import IntegrityError
+from urlparse import urlsplit, urlunsplit
+
 
 class Class():
     
@@ -68,42 +71,80 @@ class UrlTest(TestCase):
 
     def request_redirect(self, path, data={}, ajax=False,
                          view_name=None, expected_url=None,
-                         status_code=302, target_status_code=200):
+                         status_code=302, form=None):
         response = self.request(path, data, ajax)
         self.verify_view(response, view_name)
-        self.assertRedirects(response, expected_url, status_code,
-                             target_status_code)
+        if form:
+            self.assert_form_redirect(response,
+                                      expected_url,
+                                      form, status_code)
+        else:
+            self.assert_redirect_no_follow(response,
+                                           expected_url,
+                                           status_code)
         return response
 
+    def assert_redirect_no_follow(self, response, expected_url,
+                    status_code = 302, host=None, msg_prefix=''):
+        if msg_prefix:
+            msg_prefix += ": "
+        self.assertEqual(response.status_code, status_code,
+                msg_prefix + "Response didn't redirect as expected: Response"
+                " code was %d (expected %d)" %
+                    (response.status_code, status_code))
+        url = response['Location']
+        e_scheme, e_netloc, e_path, e_query, e_fragment = urlsplit(
+                                                              expected_url)
+        if not (e_scheme or e_netloc):
+            expected_url = urlunsplit(('http', host or 'testserver', e_path,
+                e_query, e_fragment))
+
+        self.assertEqual(url, expected_url,
+            msg_prefix + "Response redirected to '%s', expected '%s'" %
+                (url, expected_url))
+
+    def assert_form_redirect(self, response, expected_url, form = 'form',
+                             status_code = 302):
+        if response.status_code == 200:
+            context_form = response.context_data[form]
+            msgs = ['Response did not redirect as expected']
+            for field, error in context_form.errors.iteritems():
+                msgs.append('%s: %s' % (field, error.as_text()))
+            self.fail('\n'.join(msgs))
+        else:
+            self.assert_redirect_no_follow(response, expected_url, status_code)
+        
 class AdminTest(UrlTest):
-    
+
     @property
     def app(self):
         return self.__class__.app
-    
+
     @property
     def base_url(self):
         return '/admin/%s' % self.app
-    
-    def change_list_test(self, model):
+
+    def change_list_test(self, model, **kwargs):
         templates = ['admin/change_list.html',
                      'admin/base_site.html',
                      'admin/base.html']
         path = '%s/%s/' % (self.base_url, model)
         view_name = 'changelist_view'
-        return self.request_response(path, templates=templates, view_name=view_name)
-        
-    def change_form_test(self, model, instance_id):
+        return self.request_response(path, templates=templates,
+                                     view_name=view_name, **kwargs)
+
+    def change_form_test(self, model, instance_id, **kwargs):
         templates = ['admin/change_form.html', 
                      'admin/base_site.html', 
                      'admin/base.html',
                      'admin/includes/fieldset.html']
         path = '%s/%s/%s/' % (self.base_url, model, instance_id)
         view_name = 'change_view'
-        return self.request_response(path, templates=templates, view_name=view_name)
+        return self.request_response(path, templates=templates,
+                                     view_name=view_name, **kwargs)
 
 class ViewTest(FactoryTest):
-    
+
     def request(self, path, view, data={}, ajax=False,
                 credentials = None, arguments = {}):
         method = 'post' if data or ajax else 'get'
@@ -119,16 +160,16 @@ class ViewTest(FactoryTest):
         else:
             request.user = AnonymousUser()
         return view(request, **arguments)
-    
+
     def verify_response(self, response, contain_list = [],
                         not_contain_list=[], status_code=200):
-        
+
         self.assertEqual(response.status_code, status_code)
         for contain in contain_list:
             self.assertContains(response, **contain)
         for not_contain in not_contain_list:
             self.assertNotContains(response, **not_contain)
-            
+
     def request_response(self, path, view, data={}, ajax=False,
                          credentials = None, arguments = {},
                          contain_list=[], not_contain_list=[],
@@ -138,15 +179,15 @@ class ViewTest(FactoryTest):
         self.verify_response(response, contain_list,
                              not_contain_list, status_code)
         return response
-    
+
 class AjaxTest(ViewTest):
-        
+
     def request(self, path, view, data={},
                 credentials = None, arguments = {}):
         view = view[0] if hasattr(view,'__iter__') else view
         return super(AjaxTest, self).request(path, view, data,
                     True, credentials, arguments)
-    
+
     def verify_dajax_response(self, json_response_string, dajax_data):
         json_response = simplejson.loads(json_response_string)
         self.assertEqual(len(dajax_data), len(json_response))
@@ -162,13 +203,15 @@ class AjaxTest(ViewTest):
         return response
 
 class ModelsTest(TestCase):
-    
+
     def verify_natural_keys(self, models):
         for Model in models:
             obj = Model.objects.all()[0]
-            self.verify_natural_key_for_object(obj, Model)
-    
-    def verify_natural_key_for_object(self, obj, Model):
-        nk = obj.natural_key()
-        self.assertEqual(Model.objects.get_by_natural_key(*nk), obj)
+            self.verify_natural_key_for_object(obj)
 
+    def verify_natural_key_for_object(self, obj):
+        nk = obj.natural_key()
+        Model = obj.__class__
+        self.assertEqual(Model.objects.get_by_natural_key(*nk), obj)
+        obj.pk=None
+        self.assertRaises(IntegrityError, obj.save)
